"""Miscellaneous utility functions."""

import jax
import jax.numpy as jnp
import nonlinear_benchmarks
import numpy as np

from ._data_manager import InputOutputData, create_data_object
from ._model_structures import ModelBLA, ModelNonlinearLFR
from ._solve import SolveResult


def load_and_preprocess_silverbox_data() -> InputOutputData:
    """Load (from `nonlinear_benchmarks`) and preprocesses the Silverbox dataset.
    
    Returns
    -------
    data : `InputOutputData`
        Preprocessed Silverbox training data.

    """
    train = nonlinear_benchmarks.Silverbox()[0]
    u, y = train.u, train.y

    N = 8192  # number of samples per period
    R = 6  # number of random phase multisine realizations
    P = 1  # number of periods

    nu, ny = 1, 1  # SISO system

    fs = 1e7 / 2**14  # [Hz]
    f_idx = np.arange(1, 2 * 1342, 2)  # excited odd harmonics

    # Process data
    N_init = 164  # number of initial samples to be discarded
    N_z = 100  # number of zero samples separating the blocks visually
    N_tr = 400  # number of transient samples

    u_train = np.zeros((N, R))
    y_train = np.zeros((N, R))
    for k in range(R):
        if k == 0:
            u = u[N_init:]
            y = y[N_init:]
        else:
            idx = N_z + N_tr
            u = u[idx:]
            y = y[idx:]

        u_train[:, k] = u[:N]
        y_train[:, k] = y[:N]

        u = u[N:]
        y = y[N:]

    # Reshape data to required dimensions
    u_train = u_train.reshape(N, nu, R, P)
    y_train = y_train.reshape(N, ny, R, P)

    return create_data_object(u_train, y_train, f_idx, fs)


def extend_signal(u: jnp.ndarray, offset: int) -> jnp.ndarray:
    """Extend input signal by prepending last `offset` samples.

    Parameters
    ----------
    u : jnp.ndarray, shape (N, nu, R)
        Input signal.
    offset : int
        Number of samples to prepend.

    Returns
    -------
    u_ext : jnp.ndarray, shape (N + offset, nu, R)
        Extended input signal.

    """
    N = u.shape[0]
    repeats = (offset // N) + 1
    remainder = offset % N
    
    u_ext = jnp.tile(u, (repeats, 1, 1))
    if remainder > 0:
        u_ext = jnp.concatenate((u[-remainder:, ...], u_ext), axis=0)
    
    return u_ext


def compute_steady_state_bla_state(bla: ModelBLA, data: InputOutputData) -> jnp.ndarray:
    """Compute the steady-state BLA state trajectories."""
    N, nu = data.time.u.shape[:2]
    nx = bla.A.shape[0]
    
    G_xu = ModelBLA(  # parametric u->x frequency response; not the true BLA
        A=bla.A,
        B_u=bla.B_u,
        C_y=jnp.eye(nx),
        D_yu=jnp.zeros((nx, nu)), 
        ts=bla.ts,
        norm=bla.norm,
    )._frequency_response(data.freq.f)  # shape (N//2 + 1, nx, nu)
    X_bla = G_xu @ data.freq.U  # shape (N//2 + 1, nx, R)
    x_bla = jnp.fft.irfft(X_bla, n=N, axis=0)  # shape (N, nx, R)
    return x_bla


def evaluate_model_performance(
    model: ModelBLA | ModelNonlinearLFR,
    data: InputOutputData,
    *,
    x0: jnp.ndarray,
    offset: int,
    solve_result: SolveResult | None = None
) -> None:
    """Simulate `model` in the time domain and evaluate performance.

    Parameters
    ----------
    model : `ModelBLA` or `ModelNonlinearLFR`
        Model to be evaluated.
    data : `InputOutputData`
        Measured input-output data.
    x0 : jnp.ndarray of shape (nx, R)
        Initial state used at the beginning of each simulated realization.
    offset : int
        A non-negative number of initial samples to prepend to the input signal in
        order to let the system reach steady state before the main simulation
        begins. These prepended samples are discarded from the returned outputs.

        The value of `offset` must be consistent with `x0`: the provided initial
        state `x0` must correspond to the  state at time index `-offset`.
    solve_result : `SolveResult`, optional
        Result of the model solving process, containing timing and loss
        information.
        
    Raises
    ------
    TypeError
        If `model` is not of type `ModelBLA` or `ModelNonlinearLFR`.

    """
    if not isinstance(model, ModelBLA | ModelNonlinearLFR):
        raise TypeError("`model` must be either `ModelBLA` or `ModelNonlinearLFR`.")

    u, y = data.time.u, data.time.y
    N, ny, R = y.shape
        
    if offset > 0:
        u = extend_signal(u, offset)
        
    # Simulate model
    y_sim = model._simulate(u, x0)[0]
    y_sim = y_sim[offset:, ...]  # discard offset samples

    # Compute NRMSE per output channel
    error = y - y_sim
    mse = np.mean(error**2, axis=(0, 2))
    norm = np.mean(y**2, axis=(0, 2))
    nrmse = 100 * np.sqrt(mse / norm)  # as a percentage

    if solve_result is not None:
        avg_time = solve_result.iter_times.mean()
        unit = "s"
        if avg_time < 1:
            avg_time *= 1000
            unit = "ms"

        print(
            f"Optimization completed in {solve_result.wall_time:.2f}s "
            f"({solve_result.iter_count} iterations, "
            f"{avg_time:.2f}{unit}/iter).\n"
        )

    name = "NL-LFR" if isinstance(model, ModelNonlinearLFR) else "BLA"

    if ny == 1:
        print(f"{name} simulation error: {nrmse[0]:.2f}%.")
    else:
        print(f"{name} simulation error:")
        for k in range(ny):
            print(f"    output {k + 1}: {nrmse[k]:.2f}%.")
    print("")


def get_key(seed: int, tag: str) -> jax.Array:
    """Generate a deterministic key from a base seed and a tag.
    
    Parameters
    ----------
    seed : int
        Base seed.
    tag : str
        Tag to differentiate keys.
        
    Returns
    -------
    key : jax.Array
        A new random key that is a deterministic function of the inputs.

    """
    tag = hash(tag) & 0xFFFFFFFF  # ensure it's in 32-bit range
    return jax.random.fold_in(jax.random.key(seed), tag)


def real_valued(loss: jnp.ndarray) -> tuple[jnp.ndarray, jnp.ndarray]:
    """Split complex loss into real and imaginary parts."""
    return loss.real, loss.imag


def scalar_valued(loss: jnp.ndarray) -> float:
    """Compute scalar loss from complex loss by summing squared magnitudes."""
    return jnp.sum(jnp.abs(loss) ** 2)



"""NL-LFR inference and learning, optimization, and instantiation."""
import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np
import optimistix as optx

from . import _misc
from ._config import PRINT_EVERY, SEED, SOLVER, DeviceLike
from ._data_manager import FrequencyData, InputOutputData
from ._model_structures import ModelBLA, ModelNonlinearLFR
from ._solve import SolveResult, solve
from .static._feature_maps import AbstractFeatureMap
from .static._nonlin_funcs import AbstractNonlinearFunction, BasisFunctionModel


# When changing these constants, also update the corresponding docstrings!
EPSILON = 1e-8
FIXED_POINT_ITERS = 10
LAMBDA_W = 1e-2
MAX_ITER_INFERENCE_AND_LEARNING = 1000
MAX_ITER_OPTIMIZATION = 100
SIGMA = 1e-4


class DecVarsInferenceLearning(eqx.Module):
    """Decision variables for inference and learning."""

    B_w_star: jnp.ndarray = eqx.field(converter=jnp.asarray)
    C_z_star: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_yw_star: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_zu_star: jnp.ndarray = eqx.field(converter=jnp.asarray)
    
        
class ArgsInferenceLearning(eqx.Module):
    """Static arguments for inference and learning optimization."""
    
    theta_uy: tuple  # (A, B_u, C_y)
    phi: AbstractFeatureMap
    lambda_w: float
    fixed_point_iters: int
    freq_data: tuple  # (frequencies, sampling frequency, U, Y, G_yu)
    Lambda: jnp.ndarray  # shape (N//2 + 1, ny, ny)
    Tz_inv: jnp.ndarray  # shape (nz, nz)
    epsilon: float
    N: int
    recompute_fixed_point: bool


class ArgsNonlinearOptimization(eqx.Module):
    """Static arguments for nonlinear LFR optimization."""
    
    theta_static: ModelNonlinearLFR
    u: jnp.ndarray  # shape (N, nu, R)
    Y: jnp.ndarray  # shape (N//2 + 1, ny, R)
    Lambda: jnp.ndarray  # shape (N//2 + 1, ny, ny)
    x0: jnp.ndarray  # shape (nx, R)
    offset: int


### Public functions ###


def inference_and_learning(
    bla: ModelBLA,
    data: InputOutputData,
    phi: AbstractFeatureMap,
    nw: int,
    *,
    lambda_w: float = LAMBDA_W,
    fixed_point_iters: int = FIXED_POINT_ITERS,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser = SOLVER,
    freq_weighting: bool = True,
    max_iter: int = MAX_ITER_INFERENCE_AND_LEARNING,
    print_every: int = PRINT_EVERY,
    return_solve_details: bool = False,
    seed: int = SEED,
    epsilon: float = EPSILON,
    recompute_fixed_point: bool = True,
    device: DeviceLike = None
) -> ModelNonlinearLFR | tuple[ModelNonlinearLFR, SolveResult]:
    """Perform inference and learning.

    Parameters
    ----------
    bla : `ModelBLA`
        Parametric BLA model; is kept fixed during optimization.
    data : `InputOutputData`
        Measured input-output data object.
    phi : `AbstractFeatureMap`
        Any feature mapping that is linear in the parameters.
    nw : int
        Dimension of the latent signal `w`.
    lambda_w : float
        Regularization weight that controls the solution variance of `w`
        during inference. Defaults to `1e-2`.
    fixed_point_iters : int
        Number of fixed-point iterations for internal consistency. Defaults to `10`.
    solver : `optx.AbstractLeastSquaresSolver` or `optx.AbstractMinimiser`
        Any least-squares solver or general minimization solver from the
        Optimistix or Optax libraries. Defaults to 
        `optx.LevenbergMarquardt(rtol=1e-3, atol=1e-6)`.
    freq_weighting : bool
        Whether to use frequency weighting based on the inverse of the output noise 
        variance estimate. Defaults to `True`.
    max_iter : int
        Maximum number of optimization iterations. Defaults to `1000`.
    print_every : int
        Frequency of printing iteration information. If set to `0`, only a
        summary is printed. If set to `-1`, no printing is done. Defaults to `1`.
    return_solve_details : bool
        Whether to return detailed information about the optimization process. This is
        useful for e.g. plotting convergence curves. Defaults to `False`.
    seed : int
        Random seed for parameter initialization. Defaults to `42`.
    epsilon : float
        Numerical regularization constant for matrix inversion. Defaults to `1e-8`.
    recompute_fixed_point : bool
        Whether to recompute the fixed-point iterations with during the backward pass.
        Doing so increases computation time but reduces memory consumption, which can 
        be useful for large-scale problems. Defaults to `True`.
    device : `DeviceLike`, optional
        Device on which to perform the computations. Can be either a device
        name (`"cpu"`, `"gpu"`, or `"tpu"`) or a specific JAX device. If not
        provided, the default JAX device is used.
        
    Returns
    -------
    `ModelNonlinearLFR`
        Fully initialized NL-LFR model.
    `SolveResult`, optional
        More details about the optimization process, only returned if
        `return_solve_details` is `True`.

    """
    logging_enabled = print_every != -1
    if logging_enabled:
        header = " NL-LFR inference and learning "
        print(f"{header:=^72}")
        
    freq_weighting = _validate_weighting(
        freq_weighting, data.freq.Y_var_noise, logging_enabled
    )

    theta0, args = _prepare_inference_and_learning(
        bla, data, phi, nw, lambda_w, fixed_point_iters,
        seed, epsilon, recompute_fixed_point, freq_weighting
    )

    # Optimize the model parameters
    if logging_enabled:
        print("Starting iterative optimization...")
        if print_every > 0:
            bla_loss = _loss_bla(bla, data.freq, freq_weighting)
            print(f"    BLA loss: {bla_loss:.4e}")

    solve_result = solve(
        theta0, solver, args, _loss_inference_and_learning, 
        max_iter, print_every, device
    )

    scalar_loss, beta = solve_result.aux

    model = ModelNonlinearLFR(
        A=args.theta_uy[0],
        B_u=args.theta_uy[1],
        C_y=args.theta_uy[2],
        D_yu=bla.D_yu,
        B_w=solve_result.theta.B_w_star,
        C_z=args.Tz_inv @ solve_result.theta.C_z_star,
        D_yw=solve_result.theta.D_yw_star,
        D_zu=args.Tz_inv @ solve_result.theta.D_zu_star,
        func_static=_create_basis_function_model_given_beta(nw, phi, beta),
        ts=data.time.ts,
        norm=bla.norm
    )

    if logging_enabled:
        offset = data.time.u.shape[0]
        x_bla = _misc.compute_steady_state_bla_state(bla, data)
        _misc.evaluate_model_performance(
            model, data, x0=x_bla[-offset, :, :],
            offset=offset, solve_result=solve_result
        )
        
        # Compute recursive loss of the NL-LFR model to check consistency with the
        # inference and learning solution. We prepend a single period and start from
        # the corresponding steady-state BLA state.
        u_ext = _misc.extend_signal(data.time.u, offset)
        y_hat = model._simulate(u_ext, x0=x_bla[-offset, :, :])[0]
        y_hat = y_hat[offset:, ...]  # discard offset samples
        Y_hat = jnp.fft.rfft(y_hat, axis=0)
        loss_recursive = _loss_output_spectrum(data.freq.Y, Y_hat, args.Lambda)
        scalar_loss_recursive = _misc.scalar_valued(loss_recursive)

        print("NL-LFR loss consistency check:")
        print(f"    Inference & learning loss = {scalar_loss:.4e}")
        print(f"    Recursive simulation loss = {scalar_loss_recursive:.4e}\n")

    if return_solve_details:
        return model, solve_result
    return model


def optimize(
    model: ModelNonlinearLFR,
    data: InputOutputData,
    *,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser = SOLVER,
    freq_weighting: bool = True,
    max_iter: int = MAX_ITER_OPTIMIZATION,
    print_every: int = PRINT_EVERY,
    return_solve_details: bool = False,
    offset: int | None = None,
    device: DeviceLike = None
) -> ModelNonlinearLFR | tuple[ModelNonlinearLFR, SolveResult]:
    """Refine the parameters of an NL-LFR model using time-domain simulations.

    Parameters
    ----------
    model : `ModelNonlinearLFR`
        Initial NL-LFR model to optimize.
    data : `InputOutputData`
        Measured input-output data object.
    solver : `optx.AbstractLeastSquaresSolver` or `optx.AbstractMinimiser`
        Any least-squares solver or general minimization solver from the
        Optimistix or Optax libraries. Defaults to 
        `optx.LevenbergMarquardt(rtol=1e-3, atol=1e-6)`.
    freq_weighting : bool
        Whether to use frequency weighting based on the inverse of the output noise 
        variance estimate. Defaults to `True`.
    max_iter : int
        Maximum number of optimization iterations. Defaults to `100`.
    print_every : int
        Frequency of printing iteration information. If set to `0`, only a
        summary is printed. If set to `-1`, no printing is done. Defaults to `1`.
    return_solve_details : bool
        Whether to return detailed information about the optimization process. This is
        useful for e.g. plotting convergence curves. Defaults to `False`.
    offset : int, optional
        A non-negative integer which is used to select the unknown initial state of 
        the time-domain simulations. Specifically, the initial state is selected from 
        the steady-state BLA state simulations as `x0 = x_bla[-offset, :, :]`. The 
        number of input samples are prepended accordingly. This warm-up should ensure
        the system reaches steady-state before the main period begins, allowing for 
        leakage-free DFT computations (the prepended samples are later discarded). 
        Defaults to 10% of the data length if not provided.
    device : `DeviceLike`, optional
        Device on which to perform the computations. Can be either a device
        name (`"cpu"`, `"gpu"`, or `"tpu"`) or a specific JAX device. If not
        provided, the default JAX device is used.

    Returns
    -------
    `ModelNonlinearLFR`
        NL-LFR model with optimized parameters.
    `SolveResult`, optional
        More details about the optimization process, only returned if
        `return_solve_details` is `True`.

    """
    logging_enabled = print_every != -1
    if logging_enabled:
        header = " NL-LFR optimization "
        print(f"{header:=^72}")
        
    freq_weighting = _validate_weighting(
        freq_weighting, data.freq.Y_var_noise, logging_enabled
    )

    if offset is None:  # we start 10% "ahead of time"
        offset = int(np.ceil(0.1 * data.time.u.shape[0]))

    theta0, args = _prepare_nonlin_optimization(data, model, offset, freq_weighting)
 
    # Optimize the model parameters
    if logging_enabled:
        print("Starting iterative optimization...")
        if print_every > 0:
            bla_loss = _loss_bla(model._bla, data.freq, freq_weighting)
            print(f"    BLA loss: {bla_loss:.4e}")

    solve_result = solve(
        theta0, solver, args, _loss_nonlin_optimization, max_iter, print_every, device
    )

    # Combine optimized dynamic parameters with static ones
    model_opti = eqx.combine(solve_result.theta, args.theta_static)
    
    # Add original BLA to optimized model
    model_opti = eqx.tree_at(lambda tree: tree._bla, model_opti, replace=model._bla)
    
    if logging_enabled:
        _misc.evaluate_model_performance(
            model_opti, data, x0=args.x0, offset=offset, solve_result=solve_result
        )

    if return_solve_details:
        return model_opti, solve_result
    return model_opti


def connect(
    bla: ModelBLA,
    func_static: AbstractNonlinearFunction,
    sigma: float = SIGMA
) -> ModelNonlinearLFR:
    """Instantiate an NL-LFR model given the BLA and a static nonlinear function.

    Parameters
    ----------
    bla : `ModelBLA`
        Parametric BLA model.
    func_static : `AbstractNonlinearFunction`
        Nonlinear function object that defines the static relation between the latent 
        signals `z` and `w`. Note that the `seed` attribute of `func_static` is used
        to randomly initialize the matrices `B_w`, `C_z`, `D_yw`, and `D_zu`.
    sigma : float
        Standard deviation of the random values used to initialize `B_w` and `D_yw`.  
        Choosing a small value ensures the initial NL-LFR loss remains close to the  
        corresponding BLA loss. However, setting it too small reduces gradient 
        magnitude and can slow down or stall learning. Defaults to `1e-4`.

    Returns
    -------
    `ModelNonlinearLFR`
        NL-LFR model with (partly random) initial parameters.

    """
    nw, nz = func_static.nw, func_static.nz
    ny, nu = bla.D_yu.shape
    nx = bla.A.shape[0]

    key = _misc.get_key(func_static.seed, "connect")

    key_B_w, key_C_z, key_D_yw, key_D_zu = jax.random.split(key, 4)
    return ModelNonlinearLFR(
        A=bla.A,
        B_u=bla.B_u,
        C_y=bla.C_y,
        D_yu=bla.D_yu,
        B_w=sigma * jax.random.normal(key_B_w, (nx, nw)),
        C_z=jax.random.normal(key_C_z, (nz, nx)),
        D_yw=sigma * jax.random.normal(key_D_yw, (ny, nw)),
        D_zu=jax.random.normal(key_D_zu, (nz, nu)),
        func_static=func_static,
        ts=bla.ts,
        norm=bla.norm
    )


### Internal helpers ###


def _loss_output_spectrum(
    Y: jnp.ndarray,
    Y_hat: jnp.ndarray,
    Lambda: jnp.ndarray
) -> jnp.ndarray:
    """Compute the weighted loss between the measured and simulated output spectra.

    Parameters
    ----------
    Y : jnp.ndarray, shape (N//2 + 1, ny, R)
        Measured output spectrum, averaged over periods.
    Y_hat : jnp.ndarray, shape (N//2 + 1, ny, R)
        Simulated output spectrum.
    Lambda : jnp.ndarray, shape (N//2 + 1, ny, ny)
        Weight matrix.

    """
    F_tot, _, R = Y.shape
    return jnp.sqrt(Lambda / (F_tot * R)) @ (Y - Y_hat)


def _prepare_inference_and_learning(
    bla: ModelBLA,
    data: InputOutputData,
    phi: AbstractFeatureMap,
    nw: int,
    lambda_w: float,
    fixed_point_iters: int,
    seed: int,
    epsilon: float,
    recompute_fixed_point: bool,
    freq_weighting: bool
) -> tuple[ModelNonlinearLFR, ArgsInferenceLearning]:
    """Prepare initial guess and function arguments for inference and learning."""
    nz = phi.nz
    ny, nx = bla.C_y.shape
    N, nu = data.time.u.shape[:2]

    freqs = data.freq.f
    U = data.freq.U
    Y = data.freq.Y

    # Initialize theta_wz
    key = jax.random.key(seed)
    key_B_w, key_C_z, key_D_yw, key_D_zu = jax.random.split(key, 4)

    B_w_star = jax.random.normal(key_B_w, (nx, nw))
    C_z_star = jax.random.normal(key_C_z, (nz, nx))
    D_zu_star = jax.random.normal(key_D_zu, (nz, nu))
    D_yw_star = jax.random.normal(key_D_yw, (ny, nw))

    theta_wz = DecVarsInferenceLearning(B_w_star, C_z_star, D_yw_star, D_zu_star)
    theta_uy = (bla.A, bla.B_u, bla.C_y)

    # Compute z_star normalization
    G_zu = ModelBLA(  # parametric u->z frequency response; not the true BLA
        A=bla.A, B_u=bla.B_u, C_y=C_z_star, D_yu=D_zu_star, ts=bla.ts, norm=bla.norm
    )._frequency_response(freqs)

    Z_star = G_zu @ U
    z_star = np.fft.irfft(Z_star, axis=0)
    z_star_min, z_star_max = z_star.min(axis=(0, 2)), z_star.max(axis=(0, 2))
    Tz_inv = jnp.diag(2 / (z_star_max - z_star_min))

    G_yu = bla._frequency_response(freqs)
    f_data = (freqs, 1 / data.time.ts, U, Y, G_yu)

    args = ArgsInferenceLearning(
        theta_uy=theta_uy,
        phi=phi,
        lambda_w=lambda_w,
        fixed_point_iters=fixed_point_iters,
        freq_data=f_data,
        Lambda=_compute_weighting_matrix(data.freq, freq_weighting),
        Tz_inv=Tz_inv,
        epsilon=epsilon,
        N=N,
        recompute_fixed_point=recompute_fixed_point,
    )
    return theta_wz, args


def _loss_inference_and_learning(
    theta: DecVarsInferenceLearning,
    args: ArgsInferenceLearning
) -> tuple:
    """Implement the inference and learning method and compute the loss."""
    freqs, fs, U, Y, G_yu = args.freq_data

    A = args.theta_uy[0]
    B_u = args.theta_uy[1]
    C_y = args.theta_uy[2].astype(complex)

    B_w = theta.B_w_star.astype(complex)
    C_z = (args.Tz_inv @ theta.C_z_star).astype(complex)
    D_yw = theta.D_yw_star
    D_zu = args.Tz_inv @ theta.D_zu_star

    nw = D_yw.shape[1]
    nz, nx = C_z.shape
    F_tot, nu, R = U.shape

    Theta = jnp.vstack((B_w, D_yw)).T @ jnp.vstack((B_w, D_yw))
    Theta += args.epsilon / args.lambda_w * jnp.eye(nw)

    z = 2 * jnp.pi * freqs / fs
    zj = jnp.exp(z * 1j)

    I_nx = jnp.eye(nx)

    def _compute_parametric_Gs(k):
        G_x = jnp.linalg.solve(zj[k] * I_nx - A, jnp.hstack((B_u, B_w)))
        return (
            C_y @ G_x[:, nu:] + D_yw,  # G_yw
            C_z @ G_x[:, :nu] + D_zu,  # G_zu
            C_z @ G_x[:, nu:],  # G_zw
        )

    G_yw, G_zu, G_zw = jax.vmap(_compute_parametric_Gs)(jnp.arange(F_tot))

    # 1) Nonparametric inference
    def _infer_nonparametric_signals(k):
        Psi = G_yw[k, ...].T @ args.Lambda[k, ...]
        Phi = Psi @ G_yw[k, ...] + args.lambda_w * Theta
        W_hat = jnp.linalg.solve(Phi, Psi @ (Y[k, ...] - G_yu[k, ...] @ U[k, ...]))
        Z_hat = G_zu[k, ...] @ U[k, ...] + G_zw[k, ...] @ W_hat
        return W_hat, Z_hat

    W_star, Z_star = jax.vmap(_infer_nonparametric_signals)(jnp.arange(F_tot))

    # 2) Parametric learning
    w_star = jnp.fft.irfft(W_star, n=args.N, axis=0)
    z_star = jnp.fft.irfft(Z_star, n=args.N, axis=0)

    w_star_stacked = jnp.transpose(w_star, (2, 0, 1)).reshape(args.N * R, nw)
    z_star_stacked = jnp.transpose(z_star, (2, 0, 1)).reshape(args.N * R, nz)

    # 2a) Compute beta
    phi_z_star = args.phi._compute_features(z_star_stacked)
    beta_hat = jnp.linalg.solve(
        phi_z_star.T @ phi_z_star, phi_z_star.T @ w_star_stacked
    )

    # 2b) Perform fixed-point iterations
    def _fixed_point_iteration(_, phi_z):
        w_stacked = phi_z @ beta_hat
        w = jnp.transpose(w_stacked.reshape(R, args.N, nw), (1, 2, 0))
        W = jnp.fft.rfft(w, axis=0)
        Z = G_zu @ U + G_zw @ W
        z = jnp.fft.irfft(Z, n=args.N, axis=0)
        z_stacked = jnp.transpose(z, (2, 0, 1)).reshape(args.N * R, nz)
        return args.phi._compute_features(z_stacked)
    
    fixed_point_fn = (
        jax.checkpoint(_fixed_point_iteration)
        if args.recompute_fixed_point
        else _fixed_point_iteration
    )

    phi_z = jax.lax.fori_loop(
        0,
        args.fixed_point_iters,
        fixed_point_fn,
        phi_z_star,
        unroll=True
    )

    w_hat_stacked = phi_z @ beta_hat
    w_hat = jnp.transpose(w_hat_stacked.reshape(R, args.N, nw), (1, 2, 0))
    W_beta = jnp.fft.rfft(w_hat, axis=0)

    # 2c) Compute loss
    Y_hat = G_yu @ U + G_yw @ W_beta
    loss = _loss_output_spectrum(Y, Y_hat, args.Lambda)
    return _misc.real_valued(loss), (_misc.scalar_valued(loss), beta_hat)


def _prepare_nonlin_optimization(
    data: InputOutputData,
    model: ModelNonlinearLFR,
    offset: int,
    freq_weighting: bool = True
) -> tuple[ModelNonlinearLFR, ArgsNonlinearOptimization, ModelNonlinearLFR]:
    """Prepare initial guess and function arguments for nonlinear optimization."""    
    # Separate static and dynamic parameters
    theta0, theta_static = eqx.partition(model, eqx.is_inexact_array)
    
    # Ensure BLA is not part of the decision variables
    theta0 = eqx.tree_at(lambda tree: tree._bla, theta0, replace=None) 
    
    x_bla = _misc.compute_steady_state_bla_state(model._bla, data)

    args = ArgsNonlinearOptimization(
        theta_static=theta_static,
        u=_misc.extend_signal(data.time.u, offset) if offset > 0 else data.time.u, 
        Y=data.freq.Y,
        Lambda=_compute_weighting_matrix(data.freq, freq_weighting),
        x0=x_bla[-offset, :, :],
        offset=offset,
    )
    return theta0, args


def _loss_nonlin_optimization(
    theta: ModelNonlinearLFR, 
    args: ArgsNonlinearOptimization
) -> tuple:
    """Perform time-domain forward simulations and compute the loss."""
    theta = eqx.combine(theta, args.theta_static)

    # Simulate the model in time domain
    y_hat = theta._simulate(args.u, args.x0)[0]
    y_hat = y_hat[args.offset:, ...]  # discard offset samples
    Y_hat = jnp.fft.rfft(y_hat, axis=0)

    # Compute loss
    loss = _loss_output_spectrum(args.Y, Y_hat, args.Lambda)
    return _misc.real_valued(loss), (_misc.scalar_valued(loss),)

    
def _validate_weighting(
    freq_weighting: bool,
    var_noise: jnp.ndarray | None,
    print_warning: bool
) -> bool:
    """Check if weighting can be applied based on output noise variance availability."""
    if freq_weighting and var_noise is None:
        if print_warning:
            print(
                "Warning: Frequency weighting based on output noise variance "
                "requested, but such estimate is not available. Proceeding without "
                "weighting."
        )
        freq_weighting = False
    return freq_weighting


def _compute_weighting_matrix(
    freq_data: FrequencyData,
    freq_weighting: bool
) -> jnp.ndarray:
    """Compute weighting matrix containing the inverse output noise variances."""
    F_tot, ny = freq_data.Y.shape[:2]
    
    # Each (ny x ny) matrix is diagonal with elements equal to the inverse of the noise
    # variance for each output at each frequency. If this estimate is not available, no
    # weighting is applied (identity matrices are used).
    Lambda = np.zeros((F_tot, ny, ny))
    for k in range(F_tot):
        diag = np.diag(1 / freq_data.Y_var_noise[k]) if freq_weighting else np.eye(ny)
        Lambda[k, ...] = diag

    return jnp.asarray(Lambda)


def _loss_bla(bla: ModelBLA, freq_data: FrequencyData, freq_weighting: bool) -> float:
    """Compute the BLA performance on the loss function that is used in this module."""
    Lambda = _compute_weighting_matrix(freq_data, freq_weighting)
    Y = freq_data.Y
    U = freq_data.U

    Y_hat = bla._frequency_response(freq_data.f) @ U
    loss = _loss_output_spectrum(Y, Y_hat, Lambda)
    return _misc.scalar_valued(loss)


def _create_basis_function_model_given_beta(
    nw: int,
    phi: AbstractFeatureMap,
    beta: jnp.ndarray
) -> BasisFunctionModel:
    """Create a `BasisFunctionModel` instance given `beta`.

    The coefficients `beta` of a `BasisFunctionModel` are always initialized randomly 
    with a given seed. This function bypasses that behavior by immediately replacing 
    the randomly generated `beta` with the provided matrix obtained from inference and 
    learning. A dummy seed is still required for initialization, but it does not 
    affect the final result.
    """
    dummy_seed = 0
    return eqx.tree_at(
        where=lambda tree: tree.beta,
        pytree=BasisFunctionModel(nw, phi, dummy_seed),
        replace=beta,
    )






"""General-purpose wrapper around Optimistix solvers."""
from __future__ import annotations

import contextlib
import time
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

import equinox as eqx
import jax
import numpy as np
import optimistix as optx
from jaxtyping import PyTree
from optimistix._least_squares import _ToMinimiseFn
from optimistix._misc import OutAsArray

from ._config import DeviceLike


@dataclass(frozen=True)
class SolveResult:
    """Container for optimization results."""
    
    theta: PyTree
    aux: Any  # problem-dependent auxiliary output
    loss_history: np.ndarray
    iter_count: int
    iter_times: np.ndarray
    converged: bool
    wall_time: float


def tree_device_put(tree: Any, device: DeviceLike) -> Any:
    """Move only inexact-array leaves of a PyTree to the given device."""
    if device is None:
        return tree

    def _to_device(x):
        if eqx.is_inexact_array(x):
            return jax.device_put(x, device=device)
        return x

    return jax.tree_util.tree_map(_to_device, tree)


def _available_platforms() -> set[str]:
    plats: set[str] = set()
    for p in ["cpu", "gpu", "tpu"]:
        try:
            if jax.devices(p):
                plats.add(p)
        except RuntimeError:
            pass
    return plats


def _get_device(device: DeviceLike):
    if device is None:
        return None

    if isinstance(device, jax.Device):
        return device

    platform = device.lower()
    try:
        # NOTE: Calling jax.devices with a string like below is currently an
        # experimental JAX feature; the API may change.
        candidates = jax.devices(platform)
    except RuntimeError as e:
        raise RuntimeError(
            f"Requested device '{platform}', but no such device is available. "
            f"Available platforms: {sorted(_available_platforms())}"
        ) from e

    return candidates[0]


@contextlib.contextmanager
def _maybe_default_device(device):
    """Set JAX's default device inside the context if provided."""
    if device is None:
        yield
    else:
        with jax.default_device(device):
            yield


def solve(
    theta_init: PyTree,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser,
    args: tuple | eqx.Module,
    loss_fn: Callable[[PyTree, tuple | eqx.Module], tuple],
    max_iter: int,
    print_every: int,
    device: DeviceLike = None
) -> SolveResult:
    """Solve an optimization problem with an Optimistix solver."""
    jax_device = _get_device(device)
    
    with _maybe_default_device(jax_device):
        # Parameters live on the chosen device
        theta_init = tree_device_put(theta_init, jax_device)
        args = tree_device_put(args, jax_device)
        
        loss_fn = OutAsArray(loss_fn)
        
        if isinstance(solver, optx.AbstractMinimiser):
            loss_fn = _ToMinimiseFn(loss_fn)

        loss_fn = eqx.filter_closure_convert(loss_fn, theta_init, args)

        tags = frozenset()
        f_struct, aux_struct = loss_fn.out_struct
        options: dict[str, Any] = {}

        step = eqx.filter_jit(
            eqx.Partial(
                solver.step, fn=loss_fn, args=args, options=options, tags=tags
            )
        )
        terminate = eqx.filter_jit(
            eqx.Partial(
                solver.terminate, fn=loss_fn, args=args, options=options, tags=tags
            )
        )

        state = solver.init(
            loss_fn, theta_init, args, options, f_struct, aux_struct, tags
        )
        converged = terminate(y=theta_init, state=state)[0]

        iter_count = 0
        theta = theta_init
        loss_history = np.zeros((max_iter,))
        iter_times = np.zeros((max_iter,))

        _ = step(y=theta, state=state)  # warm-up JIT

        start_time = time.time()

        while not converged and iter_count < max_iter:
            iter_start = time.perf_counter()
            theta, state, aux = step(y=theta, state=state)
            iter_end = time.perf_counter()

            scalar_loss = aux[0]
            if print_every > 0 and (iter_count % print_every == 0):
                jax.debug.print(
                    "    Iter {0} | loss = {1:.4e}", iter_count + 1, scalar_loss
                )

            loss_history[iter_count] = scalar_loss
            iter_times[iter_count] = iter_end - iter_start

            converged = terminate(y=theta, state=state)[0]
            iter_count += 1

        wall_time = time.time() - start_time

    return SolveResult(
        theta=theta,
        aux=aux,
        loss_history=loss_history[:iter_count],
        iter_count=iter_count,
        iter_times=iter_times[:iter_count],
        converged=converged,
        wall_time=wall_time,
    )



"""BLA and NL-LFR model classes, optimized for use with JAX and Equinox."""

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np

from . import _misc
from ._data_manager import Normalizer
from .static._nonlin_funcs import AbstractNonlinearFunction


class ModelBLA(eqx.Module):
    """BLA model class.

    Parameters
    ----------
    A : jnp.ndarray, shape (nx, nx)
        State transition matrix.
    B_u : jnp.ndarray, shape (nx, nu)
        Input-to-state matrix.
    C_y : jnp.ndarray, shape (ny, nx)
        State-to-output matrix.
    D_yu : jnp.ndarray, shape (ny, nu)
        Input-to-output matrix.
    ts : float
        Sampling time (in seconds) of the discrete system.
    norm : Normalizer
        Contains means and standard deviations of input-output signals.

    """

    A: jnp.ndarray = eqx.field(converter=jnp.asarray)
    B_u: jnp.ndarray = eqx.field(converter=jnp.asarray)
    C_y: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_yu: jnp.ndarray = eqx.field(converter=jnp.asarray)
    ts: float
    norm: Normalizer = eqx.field(static=True)

    def _simulate(
        self,
        u: jnp.ndarray,
        x0: jnp.ndarray
    ) -> tuple[jnp.ndarray, jnp.ndarray]:
        """Simulate the BLA model in the time domain.

        To be used within an optimization loop, as it assumes normalized data.

        Parameters
        ----------
        u : jnp.ndarray, shape (N, nu, R)
            Normalized input signal.
        x0 : jnp.ndarray of shape (nx, R)
            Initial state of the system. 

        Returns
        -------
        Y : jnp.ndarray, shape (N, ny, R)
            Simulated output trajectories.
        X : jnp.ndarray, of shape (N, nx, R)
            Simulated state trajectories.
        W : jnp.ndarray, of shape (N, nw, R)
            Static nonlinear function outputs.
        Z : jnp.ndarray, of shape (N, nz, R)
            Static nonlinear function inputs.

        """
        def _make_step(k, state):
            X, Y_accum, X_accum = state
            U = jax.lax.dynamic_slice(u, (k, 0, 0), (1, nu, R)).squeeze(axis=0)

            # Model equations
            X_next = self.A @ X + self.B_u @ U
            Y = self.C_y @ X + self.D_yu @ U
            return X_next, Y_accum.at[k, ...].set(Y), X_accum.at[k, ...].set(X)

        N, nu, R = u.shape
        ny, nx = self.C_y.shape

        loop_init = (
            x0,
            jnp.zeros((N, ny, R)),  # Y_accum
            jnp.zeros((N, nx, R)),  # X_accum
        )
        Y, X = jax.lax.fori_loop(0, N, _make_step, loop_init)[1:]
        return Y, X

    def simulate(
        self,
        u: np.ndarray,
        *,
        x0: np.ndarray | None = None,
        offset: int | None = None
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Simulate the BLA model in the time domain for arbitrary input signals.

        Parameters
        ----------
        u : np.ndarray, shape (N,), (N, nu), (N, nu, R), or (N, nu, R, P)
            Input signal. This array can be 1D up to 4D, with:
            - N : number of time steps  
            - nu : number of inputs  
            - R : number of realizations  
            - P : number of periods  

            The input does not need to be normalized; this is handled internally.
            Since this is a public method (not used within an optimization loop), 
            the input also does not need to be periodic.

            If the fourth dimension P is provided, it is assumed that the input is
            periodic. In that case, all periods are internally concatenated into
            the first dimension, effectively simulating multiple periods sequentially.

        x0 : np.ndarray, shape (nx,) or (nx, R), optional
            Initial state for simulation. If not provided, the initial state is
            assumed to be zero.

        offset : int, optional
            Should only be provided if the input signal `u` is periodic. A non-negative
            integer representing the number of initial samples to prepend to the input
            signal to allow the system to reach steady-state before the main simulations
            begin. Those samples are not returned. If not provided, no samples are 
            prepended.

        Returns
        -------
        y : np.ndarray, shape (N, ny), (N, ny, R), or (N, ny, R, P)
            Simulated output time series, at least 2D, with ny as the number of
            output channels.

        t : np.ndarray, shape (N,)
            Time vector corresponding to one simulation of length N.

        x : np.ndarray, shape (N, nx), (N, nx, R), or (N, nx, R, P)
            Simulated state trajectories, at least 2D, with nx as the number of
            state variables.

        Raises
        ------
        ValueError
            If `u` has an invalid shape.
        ValueError
            If `x0` has an invalid shape.
        ValueError
            If `offset` is not a non-negative integer.

        """
        return _simulate_core(self, u, x0=x0, offset=offset, with_wz=False)

    def _frequency_response(self, f: np.ndarray) -> jnp.ndarray:
        """Compute the frequency response of the system.

        To be used within an optimization loop, as it assumes normalized data.

        Parameters
        ----------
        f : np.ndarray, shape (freqs,)
            Frequency points in Hz.

        Returns
        -------
        G : jnp.ndarray
            Frequency response matrix of shape (freqs, ny, nu).

        """
        def G(k):
            G_x = jnp.linalg.solve(zj[k] * I_nx - self.A, B_u)
            return C_y @ G_x + self.D_yu

        fs = 1 / self.ts
        z = 2 * jnp.pi * f / fs
        zj = jnp.exp(z * 1j)

        I_nx = jnp.eye(self.A.shape[0])
        B_u = self.B_u.astype(complex)  # to suppress a warning
        C_y = self.C_y.astype(complex)  # to suppress a warning
        return jax.vmap(G)(np.arange(len(f)))

    def num_parameters(self) -> int:
        """Return the total number of model parameters."""
        return self.A.size + self.B_u.size + self.C_y.size + self.D_yu.size


class ModelNonlinearLFR(ModelBLA):
    """NL-LFR model class.

    Inherits from `ModelBLA` and adds linear matrices `B_w`, `C_z`, `D_yw`, `D_zu`,
    and static nonlinear feedback.
    """

    B_w: jnp.ndarray = eqx.field(converter=jnp.asarray)
    C_z: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_yw: jnp.ndarray = eqx.field(converter=jnp.asarray)
    D_zu: jnp.ndarray = eqx.field(converter=jnp.asarray)
    func_static: AbstractNonlinearFunction
    
    # Reference to initial BLA, used for selecting initial states 
    _bla : ModelBLA = eqx.field(repr=False)
    
    def __init__(
        self,
        A: jnp.ndarray,
        B_u: jnp.ndarray,
        C_y: jnp.ndarray,
        D_yu: jnp.ndarray,
        B_w: jnp.ndarray,
        C_z: jnp.ndarray,
        D_yw: jnp.ndarray,
        D_zu: jnp.ndarray,
        func_static: AbstractNonlinearFunction,
        ts: float,
        norm: Normalizer
    ) -> None:
        """Initialize NL-LFR model.

        Parameters
        ----------
        A : jnp.ndarray, shape (nx, nx)
            State transition matrix.
        B_u : jnp.ndarray, shape (nx, nu)
            Input-to-state matrix.
        C_y : jnp.ndarray, shape (ny, nx)
            State-to-output matrix.
        D_yu : jnp.ndarray, shape (ny, nu)
            Input-to-output matrix.
        B_w : jnp.ndarray, shape (nx, nw)
            Feedback input-to-state matrix.
        C_z : jnp.ndarray, shape (nz, nx)
            State-to-static nonlinear function matrix.
        D_yw : jnp.ndarray, shape (ny, nw)
            Feedback input-to-output matrix.
        D_zu : jnp.ndarray, shape (nz, nu)
            Input-to-static nonlinear function matrix.
        func_static : AbstractNonlinearFunction
            Static nonlinear function mapping `z` to `w`.
        ts : float
            Sampling time (in seconds) of the discrete system.
        norm : Normalizer, optional
            Contains means and standard deviations of input-output signals.

        """
        super().__init__(A, B_u, C_y, D_yu, ts, norm)
        self.B_w = B_w
        self.C_z = C_z
        self.D_yw = D_yw
        self.D_zu = D_zu
        self.func_static = func_static 
        self._bla = ModelBLA(A, B_u, C_y, D_yu, ts, norm)

    def _simulate(
        self,
        u: jnp.ndarray,
        x0: jnp.ndarray
    ) -> tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
        """Simulate the NL-LFR model in the time domain.

        To be used within an optimization loop, as it assumes normalized data.

        Parameters
        ----------
        u : jnp.ndarray, shape (N, nu, R)
            Normalized input signal.
        x0 : jnp.ndarray of shape (nx, R)
            Initial state of the system.

        Returns
        -------
        Y : jnp.ndarray, shape (N, ny, R)
            Simulated output trajectories.
        X : jnp.ndarray, of shape (N, nx, R)
            Simulated state trajectories.
        W : jnp.ndarray, of shape (N, nw, R)
            Static nonlinear function outputs.
        Z : jnp.ndarray, of shape (N, nz, R)
            Static nonlinear function inputs.

        """
        def _make_step(k, state):
            X, Y_accum, X_accum, W_accum, Z_accum = state
            U = jax.lax.dynamic_slice(u, (k, 0, 0), (1, nu, R)).squeeze(axis=0)

            # Model equations
            Z = self.C_z @ X + self.D_zu @ U
            W = self.func_static._evaluate(Z.T).T
            X_next = self.A @ X + self.B_u @ U + self.B_w @ W
            Y = self.C_y @ X + self.D_yu @ U + self.D_yw @ W
            return (
                X_next,
                Y_accum.at[k, ...].set(Y),
                X_accum.at[k, ...].set(X),
                W_accum.at[k, ...].set(W),
                Z_accum.at[k, ...].set(Z),
            )

        N, nu, R = u.shape
        nz, nx = self.C_z.shape
        ny, nw = self.D_yw.shape

        loop_init = (
            x0,
            jnp.zeros((N, ny, R)),  # Y_accum
            jnp.zeros((N, nx, R)),  # X_accum
            jnp.zeros((N, nw, R)),  # W_accum
            jnp.zeros((N, nz, R)),  # Z_accum
        )
        
        Y, X, W, Z = jax.lax.fori_loop(0, N, _make_step, loop_init)[1:]
        return Y, X, W, Z

    def simulate(
        self,
        u: np.ndarray,
        *,
        x0: np.ndarray | None = None,
        offset: int | None = None
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Simulate the BLA model in the time domain for arbitrary input signals.

        Parameters
        ----------
        u : np.ndarray, shape (N,), (N, nu), (N, nu, R), or (N, nu, R, P)
            Input signal. This array can be 1D up to 4D, with:
            - N : number of time steps  
            - nu : number of inputs  
            - R : number of realizations  
            - P : number of periods  

            The input does not need to be normalized; this is handled internally.
            Since this is a public method (not used within an optimization loop), 
            the input also does not need to be periodic.

            If the fourth dimension P is provided, it is assumed that the input is
            periodic. In that case, all periods are internally concatenated into
            the first dimension, effectively simulating multiple periods sequentially.

        x0 : np.ndarray, shape (nx,) or (nx, R), optional
            Initial state for simulation. If not provided, the initial state is
            assumed to be zero.

        offset : int, optional
            Should only be provided if the input signal `u` is periodic. A non-negative
            integer representing the number of initial samples to prepend to the input
            signal to allow the system to reach steady-state before the main simulations
            begin. Those samples are not returned. If not provided, no samples are 
            prepended.

        Returns
        -------
        y : np.ndarray, shape (N, ny), (N, ny, R), or (N, ny, R, P)
            Simulated output time series, at least 2D, with ny as the number of
            output channels.

        t : np.ndarray, shape (N,)
            Time vector corresponding to one simulation of length N.

        x : np.ndarray, shape (N, nx), (N, nx, R), or (N, nx, R, P)
            Simulated state trajectories, at least 2D, with nx as the number of
            state variables.
        w : np.ndarray, shape (N, nw), (N, nw, R), or (N, nw, R, P)
            Simulated static nonlinear function outputs, at least 2D, with nw as
            the number of static nonlinear outputs.
        z : np.ndarray, shape (N, nz), (N, nz, R), or (N, nz, R, P)
            Simulated static nonlinear function inputs, at least 2D, with nz as
            the number of static nonlinear inputs.

        Raises
        ------
        ValueError
            If `u` has an invalid shape.
        ValueError
            If `x0` has an invalid shape.
        ValueError
            If `offset` is not a non-negative integer.

        """
        y, t, x, w, z = _simulate_core(self, u, x0=x0, offset=offset, with_wz=True)
        return y, t, x, w, z

    def num_parameters(self) -> int:
        """Return the total number of model parameters."""
        return (
            self.B_w.size + self.C_z.size + self.D_yw.size + self.D_zu.size
            + super().num_parameters() + self.func_static.num_parameters
        )
        
        
def _simulate_core(
    model: ModelBLA | ModelNonlinearLFR,
    u: np.ndarray,
    *,
    x0: np.ndarray | None,
    offset: int | None,
    with_wz: bool,
):
    """Simulate either a BLA or NL-LFR model for arbitrary input signals."""
    _validate_user_inputs(model, u, offset, x0)

    u_dim = u.ndim

    # Ensure `u` is 4D: (N, nu, R, P)
    u = u.reshape(u.shape + (1,) * (4 - u.ndim))
    N, nu, R, P = u.shape

    # Stack periods into the first dimension: (N * P, nu, R)
    u = jnp.transpose(u, (0, 3, 1, 2)).reshape(N * P, nu, R, order="F")

    if offset is not None:
        u = _misc.extend_signal(u, offset)

    nx = model.A.shape[0]
    x0 = jnp.zeros((nx, R)) if x0 is None else jnp.asarray(x0)

    # Normalize input
    u_mean = model.norm.u_mean.reshape(1, -1, 1)
    u_std = model.norm.u_std.reshape(1, -1, 1)
    u = (u - u_mean) / u_std

    u = jnp.asarray(u)

    # Call model-specific simulator
    if with_wz:
        y, x, w, z = model._simulate(u, x0)
    else:
        y, x = model._simulate(u, x0)
        w = z = None

    # Remove offset samples from outputs
    if offset is not None:
        y = y[offset:, ...]
        x = x[offset:, ...]
        if with_wz:
            w = w[offset:, ...]
            z = z[offset:, ...]

    # Denormalize output
    y_mean = model.norm.y_mean.reshape(1, -1, 1)
    y_std = model.norm.y_std.reshape(1, -1, 1)
    y = y * y_std + y_mean

    # Helper to reshape back to match input structure
    def _reshape_back(arr):
        arr = jnp.reshape(arr, (N, P, -1, R), order="F").transpose((0, 2, 3, 1))
        if u_dim in (1, 2):
            arr = jnp.squeeze(arr, axis=(2, 3))
        elif u_dim == 3:
            arr = jnp.squeeze(arr, axis=3)
        return arr

    y = _reshape_back(y)
    x = _reshape_back(x)
    if with_wz:
        w = _reshape_back(w)
        z = _reshape_back(z)

    t = np.arange(y.shape[0]) * model.ts

    if with_wz:
        return np.asarray(y), t, np.asarray(x), np.asarray(w), np.asarray(z)
    else:
        return np.asarray(y), t, np.asarray(x)


def _validate_user_inputs(
    model: ModelBLA | ModelNonlinearLFR,
    u: np.ndarray | jnp.ndarray,
    offset: int | None,
    x0: jnp.ndarray | None,
) -> None:
    
    if u.ndim != 1 and u.ndim != 2 and u.ndim != 3 and u.ndim != 4:
        raise ValueError(f"`u` must have 1 to 4 dimensions, got {u.ndim}D.")
    
    if x0 is not None:
        # Check if 1D or 2D
        if x0.ndim != 1 and x0.ndim != 2:
            raise ValueError(f"`x0` must be 1D or 2D, got {x0.ndim}D.")
        
        # Check consistency with `u`
        if u.ndim >= 3:
            if x0.ndim == 1:
                raise ValueError(
                    "`x0` must be 2D to match number of realizations in `u`."
                )
            if u.shape[2] != x0.shape[-1]:
                raise ValueError(
                    f"`x0` has {x0.shape[-1]} realizations, but `u` has "
                    f"{u.shape[2]} realizations."
                )
        else:
            if x0.ndim == 2:
                raise ValueError(
                    f"`x0` must be 1D since `u` is {u.ndim}D < 3D."
                )
            
        # Check consistency of state dimension
        if x0.shape[0] != model.A.shape[0]:
            raise ValueError(
                f"`x0` must have shape ({model.A.shape[0]}, ...), got {x0.shape}."
            )
    else:
        if u.ndim >= 3:
            x0 = jnp.zeros((model.A.shape[0], u.shape[2]))
        else:
            x0 = jnp.zeros((model.A.shape[0],))
            
    if offset is not None:
        if not (isinstance(offset, int) and offset >= 0):
            raise ValueError("`offset` must be a non-negative integer.")
 


"""Submodule `static`.

This submodule provides the building blocks for the static nonlinear function.

Exposed components:
- `basis`: Feature maps (mapping `z` to `features`) that are linear in the parameters.
   They can be used for inference and learning and nonlinear optimization.
- `BasisFunctionModel`: Static nonlinear function (mapping `z` to `w`) based on a
   feature map.
- `NeuralNetwork`:  Static nonlinear function (mapping `z` to `w`) based on a neural
   network.
"""


from . import _feature_maps as basis
from ._nonlin_funcs import BasisFunctionModel, NeuralNetwork


__all__ = ["basis", "BasisFunctionModel", "NeuralNetwork"]



"""Top-level package for frequency-domain system identification.

The API is structured around three main namespaces:

- ``lin``:
    Tools for creating and optimizing BLA models.

- ``nonlin``:
    Tools for creating and optimizing nonlinear LFR models.

- ``static``:
    Static nonlinear function approximators and feature-map constructions.

Additional components:
- ``create_data_object`` for constructing ``InputOutputData`` objects, the starting
    point for all identification routines.
-  ``InputOutputData``, ``TimeData``, ``FrequencyData``, ``NonparametricBLA``,
    and ``Normalizer`` data objects.
- ``ModelBLA`` and ``ModelNonlinearLFR`` as the core model classes.
- ``SolveResult`` for detailed information about optimization routines.
- ``load_and_preprocess_silverbox_data`` for loading the Silverbox example dataset. 
    This is useful for quickly loading a low-dimensional benchmark dataset in the
    required format.
"""


from . import _best_linear_approximation as lin
from . import _nonlin_lfr as nonlin
from . import static
from ._data_manager import (
    FrequencyData,
    InputOutputData,
    NonparametricBLA,
    Normalizer,
    TimeData,
    create_data_object,
)
from ._misc import load_and_preprocess_silverbox_data
from ._model_structures import ModelBLA, ModelNonlinearLFR
from ._solve import SolveResult


__all__ = [
    "lin",
    "nonlin",
    "static",
    "FrequencyData",
    "InputOutputData",
    "NonparametricBLA",
    "Normalizer",
    "TimeData",
    "create_data_object",
    "load_and_preprocess_silverbox_data",
    "ModelBLA",
    "ModelNonlinearLFR",
    "SolveResult"
]



"""Nonlinear feature mappings (`z` to `features`) that are linear in the parameters."""

from abc import abstractmethod
from itertools import combinations_with_replacement

import equinox as eqx
import jax
import jax.numpy as jnp
import numpy as np


class AbstractFeatureMap(eqx.Module):
    """Abstract base class for feature mappings.

    Subclasses must provide the attributes `nz` and `num_features`,
    and must implement the method `_compute_features()`.
    """

    nz: eqx.AbstractVar[int]
    num_features: eqx.AbstractVar[int]

    @abstractmethod
    def _compute_features(self, z: jnp.ndarray) -> jnp.ndarray:
        """Compute the nonlinear feature mapping.

        From inputs of shape (..., `nz`) to outputs of shape (..., `num_features`).
        """
        ...


class Polynomial(AbstractFeatureMap):
    """Flexible polynomial feature map.

    This class constructs and evaluates polynomial basis features up to a
    specified degree, with optional cross-terms, offset, linear terms, and
    `tanh` clipping of the inputs.
    """

    nz: int
    degree: int
    type: str
    cross_terms: bool
    offset: bool
    linear: bool
    tanh_clip: bool
    num_features: int
    combination_matrix: jnp.ndarray = eqx.field(repr=False)

    def __init__(
        self,
        nz: int,
        degree: int,
        type: str = "full",
        cross_terms: bool = True,
        offset: bool = True,
        linear: bool = True,
        tanh_clip: bool = True,
    ) -> None:
        """Initialize a flexible polynomial feature map.

        Parameters
        ----------
        nz : int
            Number of input features (dimension of latent signal `z`).
        degree : int
            Maximum polynomial degree.
        type : str, optional
            Type of polynomial. Must be one of `"full"`, `"odd"`, or `"even"`.
            Defaults to `"full"`.
        cross_terms : bool, optional
            Whether to include cross-terms in the polynomial features.
            Defaults to `True`.
        offset : bool, optional
            Whether to include a constant offset term in the features.
            Defaults to `True`.
        linear : bool, optional
            Whether to include linear terms in the polynomial features.
            Defaults to `True`.
        tanh_clip : bool, optional
            Whether to apply `tanh` clipping to the input features before
            constructing polynomial terms. Defaults to `True`.

        """
        self.nz = nz
        self.degree = degree
        self.type = type
        self.cross_terms = cross_terms
        self.offset = offset
        self.linear = linear
        self.tanh_clip = tanh_clip

        # Construct the feature architecture based on the specified parameters.
        if self.type == "full":
            active_degrees = range(1 if self.linear else 2, self.degree + 1)
        elif self.type == "odd":
            active_degrees = range(1 if self.linear else 3, self.degree + 1, 2)
        elif self.type == "even":
            active_degrees = range(2, self.degree + 1, 2)
        else:
            raise ValueError(
                'Invalid polynomial `type`. Must be "full", "odd", or "even".'
            )

        max_degree = active_degrees[-1]
        combination_matrix = []
        combination_row = np.full((max_degree,), -1, dtype=int)

        num_features = 1 if self.offset else 0
        for _, deg in enumerate(active_degrees):
            if self.cross_terms:
                combinations = tuple(
                    combinations_with_replacement(range(self.nz), deg)
                )
            else:
                combinations = tuple((i,) * deg for i in range(self.nz))

            num_features += len(combinations)
            for _, comb in enumerate(combinations):
                combination_row[:deg] = comb
                combination_matrix.append(combination_row.copy())

        self.num_features = num_features
        self.combination_matrix = jnp.array(combination_matrix)

    def _compute_features(self, z: jnp.ndarray) -> jnp.ndarray:
        N, nz = z.shape
        if nz != self.nz:
            raise ValueError(
                "Input size does not match the basis function size: "
                "`z.shape[1] != nz`."
            )

        if self.tanh_clip:
            z = jnp.tanh(z)

        # Augment input to make it consistent with the combination matrix
        z_augmented = jnp.hstack((z, jnp.ones((N, 1))))

        def _compute_phi_z(combination_idx):
            combination = self.combination_matrix[combination_idx]
            z_selected = jnp.take(z_augmented, combination.ravel(), axis=1)
            return jnp.prod(z_selected, axis=1)

        num_combs = self.combination_matrix.shape[0]
        phi_z = jax.vmap(_compute_phi_z, out_axes=1)(jnp.arange(num_combs))

        return jnp.hstack((jnp.ones((N, 1)), phi_z)) if self.offset else phi_z


class LegendrePolynomial(AbstractFeatureMap):
    """Legendre polynomial feature map.

    This class constructs and evaluates Legendre polynomial features up to a
    specified degree, with optional offset terms and optional `tanh` clipping
    of the inputs.
    """

    nz: int
    degree: int
    offset: bool
    tanh_clip: bool
    num_features: int

    def __init__(
        self,
        nz: int,
        degree: int,
        offset: bool = True,
        tanh_clip: bool = True,
    ) -> None:
        """Initialize a Legendre polynomial feature map.

        Parameters
        ----------
        nz : int
            Number of input features (dimension of latent signal `z`).
        degree : int
            Maximum polynomial degree.
        offset : bool, optional
            Whether to include a constant offset term. Defaults to `True`.
        tanh_clip : bool, optional
            Whether to apply `tanh` clipping to the inputs. Defaults to `True`.

        """
        self.nz = nz
        self.degree = degree
        self.offset = offset
        self.tanh_clip = tanh_clip
        self.num_features = self.nz * self.degree + (1 if self.offset else 0)

    def _compute_features(self, z: jnp.ndarray) -> jnp.ndarray:

        def _compute_phi_z(k, state):
            phi_z, phi_z_previous, phi_z_two_before = state
            phi_z_current = (
                (2 * k - 1) / k * z * phi_z_previous
                - (k - 1) / k * phi_z_two_before
            )
            phi_z = phi_z.at[k - 2, ...].set(phi_z_current)
            return phi_z, phi_z_current, phi_z_previous

        if self.tanh_clip:
            z = jnp.tanh(z)

        N = z.shape[0]
        phi_z0 = jnp.zeros((self.degree - 1, N, self.nz))
        loop_init = (phi_z0, z, jnp.ones_like(z))

        phi_z = jax.lax.fori_loop(
            2, self.degree + 1, _compute_phi_z, loop_init, unroll=True
        )[0]

        phi_z = jnp.transpose(phi_z, (1, 2, 0)).reshape(N, -1)
        phi_z = jnp.hstack((jnp.ones((N, 1)), z, phi_z))
        return phi_z if self.offset else phi_z[:, 1:]


class ChebyshevPolynomial(AbstractFeatureMap):
    """Chebyshev polynomial feature map.

    This class constructs and evaluates Chebyshev polynomial features of the
    first or second kind up to a specified degree, with optional offset terms
    and optional `tanh` clipping of the inputs.
    """

    nz: int
    degree: int
    type: int
    offset: bool
    tanh_clip: bool
    num_features: int

    def __init__(
        self,
        nz: int,
        degree: int,
        type: int,
        offset: bool = True,
        tanh_clip: bool = True,
    ) -> None:
        """Initialize a Chebyshev polynomial feature map.

        Parameters
        ----------
        nz : int
            Number of input features (dimension of latent signal `z`).
        degree : int
            Maximum polynomial degree.
        type : int
            Polynomial type:
            - `1`: First kind (orthogonal w.r.t. 1/sqrt(1 - x²))
            - `2`: Second kind (orthogonal w.r.t. sqrt(1 - x²))
        offset : bool, optional
            Whether to include a constant offset term. Defaults to `True`.
        tanh_clip : bool, optional
            Whether to apply `tanh` clipping to the inputs. Defaults to `True`.

        """
        if type not in (1, 2):
            raise ValueError('Invalid polynomial `type`. Must be `1` or `2`.')

        self.nz = nz
        self.degree = degree
        self.type = type
        self.offset = offset
        self.tanh_clip = tanh_clip
        self.num_features = self.nz * self.degree + (1 if self.offset else 0)

    def _compute_features(self, z: jnp.ndarray) -> jnp.ndarray:

        def _compute_phi_z(k, state):
            phi_z, phi_z_previous, phi_z_two_before = state
            phi_z_current = 2 * z * phi_z_previous - phi_z_two_before
            phi_z = phi_z.at[k - 2, ...].set(phi_z_current)
            return phi_z, phi_z_current, phi_z_previous

        if self.tanh_clip:
            z = jnp.tanh(z)

        N = z.shape[0]
        phi_z0 = jnp.zeros((self.degree - 1, N, self.nz))
        loop_init = (phi_z0, self.type * z, jnp.ones_like(z))

        phi_z = jax.lax.fori_loop(
            2, self.degree + 1, _compute_phi_z, loop_init, unroll=True
        )[0]

        phi_z = jnp.transpose(phi_z, (1, 2, 0)).reshape(N, -1)
        phi_z = jnp.hstack((jnp.ones((N, 1)), self.type * z, phi_z))
        return phi_z if self.offset else phi_z[:, 1:]



"""General static nonlinear function mappings (mapping `z` to `w`)."""

from abc import abstractmethod
from collections.abc import Callable

import equinox as eqx
import jax
import jax.numpy as jnp

from .. import _misc
from .._config import SEED
from ..static._feature_maps import AbstractFeatureMap


class AbstractNonlinearFunction(eqx.Module):
    """Abstract base class for nonlinear function mappings.

    Subclasses must provide the attributes `nw`, `nz`, `seed`, and
    `num_parameters`, and implement the method `_evaluate()`.
    """

    nw: eqx.AbstractVar[int]
    nz: eqx.AbstractVar[int]
    seed: eqx.AbstractVar[int]
    num_parameters: eqx.AbstractVar[int]

    @abstractmethod
    def _evaluate(self, z: jnp.ndarray) -> jnp.ndarray:
        """Evaluate the nonlinear function.

        From inputs of shape (..., `nz`) to outputs of shape (..., `nw`).
        """
        ...


class BasisFunctionModel(AbstractNonlinearFunction):
    """Static nonlinear function based on an `AbstractFeatureMap`.

    This class combines an `AbstractFeatureMap` with a coefficient matrix `beta`
    to implement and evaluate a static nonlinear mapping that is linear in its
    parameters.
    """

    nw: int
    nz: int
    beta: jnp.ndarray
    phi: AbstractFeatureMap
    num_parameters: int
    seed: int = eqx.field(repr=False)

    def __init__(
        self,
        nw: int,
        phi: AbstractFeatureMap,
        seed: int = SEED,
    ) -> None:
        """Initialize a static nonlinear basis-function model.

        Parameters
        ----------
        nw : int
            Number of output features (dimension of latent signal `w`).
        phi : AbstractFeatureMap
            Nonlinear feature map that is linear in the parameters.
        seed : int, optional
            Used for randomly initializing (i) the nonlinear coefficient
            matrix `beta` and (ii) the matrices `B_w`, `C_z`, `D_yw`, and
            `D_zu` (initialized externally, not by this class). Defaults
            to `42`.

        """
        self.nw = nw
        self.phi = phi
        self.nz = phi.nz
        self.seed = seed

        self.beta = jax.random.uniform(
            key=_misc.get_key(self.seed, "basis_function_model"),
            shape=(self.phi.num_features, self.nw),
            minval=-1.0,
            maxval=1.0,
        )
        self.num_parameters = self.beta.size

    def _evaluate(self, z: jnp.ndarray) -> jnp.ndarray:
        return self.phi._compute_features(z) @ self.beta


class NeuralNetwork(AbstractNonlinearFunction):
    """Fully connected feedforward neural network.

    This class wraps an `eqx.nn.MLP` and exposes a simple interface for
    evaluating the network.
    """

    nw: int
    nz: int
    model: eqx.nn.MLP
    num_parameters: int
    layers: int = eqx.field(repr=False)
    neurons_per_layer: int = eqx.field(repr=False)
    activation: Callable = eqx.field(repr=False)
    seed: int = eqx.field(repr=False)
    bias: bool = eqx.field(repr=False)
    

    def __init__(
        self,
        nz: int,
        nw: int,
        layers: int,
        neurons_per_layer: int,
        activation: Callable,
        seed: int = SEED,
        bias: bool = True,
    ) -> None:
        """Initialize a fully connected feedforward neural network.

        Parameters
        ----------
        nz : int
            Number of input features (dimension of latent signal `z`).
        nw : int
            Number of output features (dimension of latent signal `w`).
        layers : int
            Number of hidden layers.
        neurons_per_layer : int
            Number of neurons per hidden layer.
        activation : Callable, from `jax.nn`
            Activation function used in hidden layers.
        seed : int, optional
            Used for randomly initializing (i) the neural network parameters and
            (ii) the matrices `B_w`, `C_z`, `D_yw`, and `D_zu` (initialized
            externally, not by this class). Defaults to `42`.
        bias : bool, optional
            Whether to include bias terms. Defaults to `True`.

        """
        self.nz = nz
        self.nw = nw
        self.layers = layers
        self.neurons_per_layer = neurons_per_layer
        self.activation = activation
        self.seed = seed
        self.bias = bias

        self.model = eqx.nn.MLP(
            in_size=self.nz,
            out_size=self.nw,
            width_size=self.neurons_per_layer,
            depth=self.layers,
            activation=self.activation,
            use_bias=self.bias,
            key=_misc.get_key(self.seed, "neural_network"),
        )
        self.num_parameters = sum(
            x.size
            for x in jax.tree_util.tree_leaves(self.model)
            if isinstance(x, jax.Array)
        )

    def _evaluate(self, z: jnp.ndarray) -> jnp.ndarray:
        return jax.vmap(self.model)(z)
